#!/usr/bin/perl
# this is to fixup the webpal key timings and
# to provide for a KEY_DOWN sense.
# Timings were gathered from irrecord, 
# then reformatted into raw lines such as:
#
# name F6 2304 818 761 800 1548 798 753 812 763 1578 771 2355 2316
#
# the above line, after processing, would then become:
#
# name F6 2343 781 781 781 1562 781 781 781 781 1562 781 2343 2343
#
#
# this gives more accurate readings of the IR signal from the keyboard.
#

my @KeyLine;
my $Key;

sub PrintHeader {
print "
begin remote

  name   WebPal_Keyboard
  flags CONST_LENGTH|RAW_CODES
  eps            10
  aeps           50

  ptrail          0
  repeat     0     0
  gap    21872

  begin raw_codes

";
}

sub PrintFooter {
print "  end raw_codes

end remote
";
}

sub FixTimings {
my $Index = $_[0];
my $full = $KeyLine[$Index] / 781;
my $mod = $KeyLine[$Index] % 781;
my @temp = split(/\./,$full);
my $Corrected = $temp[0] * 781;
	if (abs (781 - $mod) < 100) { $Corrected += 781 };
	$KeyLine[$Index] = $Corrected;
}

sub FixLine {
	@KeyLine = split(/ /,$_[0]);
	$Key = @KeyLine[1]; $Key =~ tr/a-z/A-Z/;
	shift @KeyLine; shift @KeyLine;
	for ($i=0; $i<=$#KeyLine; $i++) { FixTimings($i) };
		# remove 8 entries of prefix from keycode.
	shift @KeyLine; shift @KeyLine;
	shift @KeyLine; shift @KeyLine;
	shift @KeyLine; shift @KeyLine;
	shift @KeyLine; shift @KeyLine;
}

sub PrintPrefix {
# print 3,1,1,1 time prefix
	print "        2343 781 781 781";
}

sub PrintMake {
# print 2,1,1,1 of keydown prefix.
	print " 1562 781 781 781\n       ";
}

sub PrintBreak {
# print 1,4 of keyup prefix.
	print " 781 3124\n       ";
}

sub PrintMakeCode {
# print keydown & keyup codes.
my $i;
	print "    name KEY_".$Key."\n";
	PrintPrefix(); PrintMake();
	for ($i=0; $i<=$#KeyLine; $i++) {
		print " ".$KeyLine[$i];
	}
	print "\n\n";
}

sub PrintBreakCode {
# print keydown & keyup codes.
my $i;
	print "    name BRK_".$Key."\n";
	PrintPrefix(); PrintBreak();
	for ($i=0; $i<=$#KeyLine; $i++) {
		print " ".$KeyLine[$i];
	}
	print "\n\n";
}

sub PrintAllUpCode {
	print "    name ALL_UP\n";
	PrintPrefix(); PrintBreak();
	print " 7029\n\n";
}

sub DumpKeyCodes {
	PrintMakeCode();
	PrintBreakCode();
}

	PrintHeader();
	while (<STDIN>) {
		chomp;
		FixLine($_);
		DumpKeyCodes();
	}
	PrintAllUpCode();
	PrintFooter();

