/*
* Copyright 2004 Yann Vernier <yann@donkey.dyndns.org>
*
* Copyright 2005 Tom Walsh <tom@openhardware.net>
* Yann had the idea, Tom made it work!  ;-)
*
*/

/* SHANNON or WEBPAL keyboard? */
//#define SHANNON	1
#define SHANNON	0

/* you got problems? */
#undef	DEBUG
//#define	DEBUG	0

 /* Standard headers */
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdbool.h>
#include <time.h>
#include <sys/times.h>
#include <syslog.h>
#include	<signal.h>

/* Input subsystem interface */
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/ioctl.h>

/* Unix socket support */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <sys/poll.h>

/* Keyboard map */
#if SHANNON!=0
#include "CodesToKeys_shannon.h"
#include "CodesToMod_shannon.h"
#else
#include "CodesToKeys_webpal.h"
#include "CodesToMod_webpal.h"
#endif

#ifdef DEBUG
#define DBG(x...) printf (x)
#else
#define DBG(x...) do {} while (0)
#endif

#define	FALSE		0
#define	TRUE		!FALSE

#define PIDFILE	"/var/run/lirc_keybd.pid"
#define progname	"lirc_keybd"

int	lirc_fd, ui_fd, fd;
char *pidfile = PIDFILE;
FILE	*pidf;
#ifdef DEBUG
FILE	*copy_fd;
#endif

// crazy math on Modifiers, even elements are valid
// odd elements are not used. see FuncX() Alt,Ctrl,etc. stuff.
unsigned short Modifiers [AllKeysUp+2];

#define  TimerIn10ms             0
#define  TimerInSeconds          1
#define	AutoDelay10MS				40
#define	AutoDeath10MS				10

typedef	unsigned long	HEX;
typedef	unsigned short	BYTE;

    /* used to run elapsed unix timers. */
HEX  TestTimerNow, TestTimeStart;

struct   TIMER {
      HEX   SetTime;    /* when it was set. */
      int   HowLong;    /* how long it is to run. */
      BYTE  Type;       /* TimerIn10ms or TimerInSeconds. */
   };

static struct TIMER   KeyTimer, RepeatTimer;

#if SHANNON!=0
const char RemoteName[]="Shannon_Keyboard";
#else
const char RemoteName[]="WebPal_Keyboard";
#endif

const struct sockaddr_un sa={
	.sun_family = AF_UNIX,
	.sun_path = "/dev/lircd"
};

char	buf[BUFSIZ];
int	size, keysdown=0;
unsigned short	KeyCode, LastKeyCode;


int CreateUinputDevice(void) {
	int fd, i;
	struct uinput_user_dev dev;
	
	fd=open("/dev/uinput", O_WRONLY);
	if(fd<0) {
			// retry on /dev/input/uinput
		fd=open("/dev/input/uinput", O_WRONLY);
		if(fd<0) {
			perror("cannot open /dev/uinput OR /dev/input/uinput");
			return fd;
		}
	}

	memset(&dev, 0, sizeof(dev));
	strncpy(dev.name, "LIRC Keyboard", UINPUT_MAX_NAME_SIZE);

	if(write(fd, &dev, sizeof(dev))<0) {
		perror("write struct uinput_user_dev");
		close(fd);
		return -1;
	}

	if(ioctl(fd,UI_SET_EVBIT,EV_KEY)==-1) {
		perror("ioctl UI_SET_EVBIT EV_KEY");
		close(fd);
		return -1;
	}

	if(ioctl(fd,UI_SET_EVBIT,EV_REP)==-1) {
		perror("ioctl UI_SET_EVBIT EV_REP");
		close(fd);
		return -1;
	}

	for(i=0; i<CODETOKEYS_SIZE; i++) {
		if(CodesToKeys[i])
			if(ioctl(fd,UI_SET_KEYBIT,CodesToKeys[i])==-1) {
			perror("ioctl UI_SET_KEYBIT");
			close(fd);
			return -1;
		}
	}

	if(ioctl(fd,UI_DEV_CREATE)==-1) {
		perror("ioctl UI_DEV_CREATE");
		close(fd);
		return -1;
	}
	return fd;
}

int DestroyUinputDevice(int fd) {
	if(ioctl(fd,UI_DEV_DESTROY)==-1) {
		perror("ioctl UI_DEV_DESTROY");
		return -1;
	}
	return close(fd);
}

int CreateLircdSocket (void)
{
	lirc_fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (lirc_fd < 0) {
		perror ("failed socket operation");
		return -1;
	}
		// set socket to non-blocking.
//	fcntl (lirc_fd, F_SETFL, fcntl (lirc_fd, F_GETFL, 0) | O_NONBLOCK);
		// keep socket at blocking.
	fcntl (lirc_fd, F_SETFL, fcntl (lirc_fd, F_GETFL, 0));
	if (connect (lirc_fd, (struct sockaddr*)&sa, sizeof(sa)) < 0) {
		perror("connect to /dev/lircd");
		close(lirc_fd);
		return -1;
	}
	return 0;
}

void SysKeypress(int fd, int code, int down) {
	struct input_event ev;
	memset (&ev, 0, sizeof(ev));
	ev.type=EV_KEY; ev.code=code; ev.value=down;
	write (fd, &ev, sizeof(ev));
}

#define	UINPUT_CREATE_FAILED		1
#define	SOCKET_CREATE_FAILED		2

int ThisIsOurRemote (void)
{// check remote name to verify it is the keyboard.
	int	name_size = sizeof (RemoteName)-1;
	char	*end;
	end=memchr (buf, '\n', size);
	if (end && !memcmp (end-name_size, RemoteName, name_size)) {
		*end=0;
		return 1;
	} else return 0;
}

void EraseModifiers (void)
{
	int	i;
	for (i=0; i<AllKeysUp+1; i++) { Modifiers [i] = 0; }
}

void FuncLeftAltDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_LEFTALT, 1);
	DBG ("FuncLeftAltDn (void)\n");
}

void FuncLeftAltUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_LEFTALT, 0);
	DBG ("FuncLeftAltUp (void)\n");
}

void FuncRightAltDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_RIGHTALT, 1);
	DBG ("FuncRightAltDn (void)\n");
}

void FuncRightAltUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_RIGHTALT, 0);
	DBG ("FuncRightAltUp (void)\n");
}

void FuncLeftCtrlDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_LEFTCTRL, 1);
	DBG ("FuncLeftCtrlDn (void)\n");
}

void FuncLeftCtrlUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_LEFTCTRL, 0);
	DBG ("FuncLeftCtrlUp (void)\n");
}

void FuncLeftShftDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_LEFTSHIFT, 1);
	DBG ("FuncLeftShftDn (void)\n");
}

void FuncLeftShftUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_LEFTSHIFT, 0);
	DBG ("FuncLeftShftUp (void)\n");
}

void FuncRightShftDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_RIGHTSHIFT, 1);
	DBG ("FuncRightShftDn (void)\n");
}

void FuncRightShftUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_RIGHTSHIFT, 0);
	DBG ("FuncRightShftUp (void)\n");
}

void FuncRightCtrlDn (unsigned short state)
{
	if (Modifiers [state & ~1]) return;
	Modifiers [state & ~1] = 1;
	SysKeypress (ui_fd, KEY_RIGHTCTRL, 1);
	DBG ("FuncRightCtrlDn (void)\n");
}

void FuncRightCtrlUp (unsigned short state)
{
	Modifiers [state & ~1] = 0;
	SysKeypress (ui_fd, KEY_RIGHTCTRL, 0);
	DBG ("FuncRightCtrlUp (void)\n");
}

void FuncAllKeysUp (unsigned short state)
{
		// use down to index array, if set, call up function.
	if (Modifiers [LeftAltDn]) {
		Modifiers [LeftAltDn] = 0;
		FuncLeftAltUp(LeftAltUp);
	}
	if (Modifiers [RightAltDn]) {
		Modifiers [RightAltDn] = 0;
		FuncRightAltUp(RightAltUp);
	}
	if (Modifiers [LeftCtrlDn]) {
		Modifiers [LeftCtrlDn] = 0;
		FuncLeftCtrlUp(LeftCtrlUp);
	}
	if (Modifiers [LeftShftDn]) {
		Modifiers [LeftShftDn] = 0;
		FuncLeftShftUp(LeftShftUp);
	}
	if (Modifiers [RightShftDn]) {
		Modifiers [RightShftDn] = 0;
		FuncRightShftUp(RightShftUp);
	}
	if (Modifiers [RightCtrlDn]) {
		Modifiers [RightCtrlDn] = 0;
		FuncRightCtrlUp(RightCtrlUp);
	}
		// just to be sure.
	EraseModifiers();
	DBG ("FuncAllKeysUp (void)\n");
}

unsigned short KeyCodeIsModifier (unsigned short Keycode)
{// table the keycode to see if special modifier keystroke.
	switch (CodesToMod [Keycode]) {
		case AllKeysUp:
			FuncAllKeysUp(AllKeysUp);
			return 1;
		case LeftAltDn:
			FuncLeftAltDn(LeftAltDn);
			return 1;
		case LeftAltUp:
			FuncLeftAltUp(LeftAltUp);
			return 1;
		case RightAltDn:
			FuncRightAltDn(RightAltDn);
			return 1;
		case RightAltUp:
			FuncRightAltUp(RightAltUp);
			return 1;
		case LeftCtrlDn:
			FuncLeftCtrlDn(LeftCtrlDn);
			return 1;
		case LeftCtrlUp:
			FuncLeftCtrlUp(LeftCtrlUp);
			return 1;
		case LeftShftDn:
			FuncLeftShftDn(LeftShftDn);
			return 1;
		case LeftShftUp:
			FuncLeftShftUp(LeftShftUp);
			return 1;
		case RightShftDn:
			FuncRightShftDn(RightShftDn);
			return 1;
		case RightShftUp:
			FuncRightShftUp(RightShftUp);
			return 1;
		case RightCtrlDn:
			FuncRightCtrlDn(RightCtrlDn);
			return 1;
		case RightCtrlUp:
			FuncRightCtrlUp(RightCtrlUp);
			return 1;
	}
	return 0;		
}

unsigned short WaitForKeyCode (void)
{
	struct pollfd pfd={
		.fd=lirc_fd,
		.events=POLLIN,
		.revents=0
	};
	unsigned long long code;
	if (poll (&pfd, 1, -1) > 0) {
		size = read (lirc_fd, buf, BUFSIZ);
		if (ThisIsOurRemote()) {
			code = strtoull (buf, NULL, 16);
			buf [size-1] = '\0';
#ifdef DEBUG
			fprintf (copy_fd, "%s\n", buf);
#endif
			DBG ("%s\n", buf);
			if (KeyCodeIsModifier(code) == 0) {
				if (strncmp (&buf[20], "BRK", 3) == 0) {
						/* any key release restarts autorepeat delay. */
					LastKeyCode = 0;
					return 0;
				}
				return (unsigned short) code;
			}
		}
	}
	return 0;
}

void create_lockfile (void)
{ /* create lockfile. */
   /* create pid lockfile in /var/run */
   if((fd=open(pidfile,O_RDWR|O_CREAT,0644))==-1 ||
      (pidf=fdopen(fd,"r+"))==NULL)
   {
      fprintf(stderr,"%s: can't open or create %s\n",
         progname,pidfile);
      perror(progname);
      exit(EXIT_FAILURE);
   }
   if(flock(fd,LOCK_EX|LOCK_NB)==-1)
   {
      pid_t otherpid;

      if(fscanf(pidf,"%d\n",&otherpid)>0)
      {
         fprintf(stderr,"%s: there seems to already be "
            "a lircd process with pid %d\n",
            progname,otherpid);
         fprintf(stderr,"%s: otherwise delete stale "
            "lockfile %s\n",progname,pidfile);
      }
      else
      {
         fprintf(stderr,"%s: invalid %s encountered\n",
            progname,pidfile);
      }
      exit(EXIT_FAILURE);
   }
   (void) fcntl(fd,F_SETFD,FD_CLOEXEC);
   rewind(pidf);
   (void) fprintf(pidf,"%d\n",getpid());
   (void) fflush(pidf);
   (void) ftruncate(fileno(pidf),ftell(pidf));
}

void DoSigTerm (int sig)
{
   (void) unlink(pidfile);
	closelog();
   signal(sig,SIG_DFL);
   raise(sig);
}

void MakeDaemon(void)
{
   if (daemon (0, 0) == -1) {
      fprintf(stderr,"daemon() failed");
      perror("daemon filed");
		DoSigTerm(SIGTERM);
   }
}


void CloseGracefully (int sig)
{
#ifdef DEBUG
	fflush (copy_fd); fclose (copy_fd);
#endif
	DestroyUinputDevice (ui_fd);
	raise(SIGTERM);
}

void SetElapsed (int HowLong, struct TIMER * timer)
{ /* set for a TimeElapsed event. */
  timer->HowLong = HowLong;
  timer->SetTime = (HEX) times (0);
}

unsigned short TimeElapsed (struct TIMER * timer)
{ /* return TRUE if preset interval has expired. */
  TestTimerNow = (HEX) times (0);
  TestTimeStart = timer->SetTime;
  if (((TestTimeStart>TestTimerNow) ? (TestTimerNow + (~TestTimeStart)) : (TestTimerNow - TestTimeStart)) > timer->HowLong) {
    return TRUE;
  }
  return FALSE;
}

int main(int argc, char *argv[]) {
		// clear ALT,SHFT,etc markers.
	EraseModifiers();
#ifdef DEBUG
	signal (SIGINT, CloseGracefully);
	if ((copy_fd = fopen("record-keys.txt", "w")) == NULL) {
		perror ("open capture");
		return -1;
	}
#endif
		// create and condition the universal input device.
	if ((ui_fd = CreateUinputDevice()) < 0) return UINPUT_CREATE_FAILED;
		// create socket to poll lircd daemon output.
	if (CreateLircdSocket () < 0) {
		DestroyUinputDevice (ui_fd);
		return SOCKET_CREATE_FAILED;
	}
#ifndef DEBUG
	signal (SIGTERM, DoSigTerm);
	MakeDaemon();
	create_lockfile ();
#endif
	openlog(progname, LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "mapping codes for %s\n", RemoteName);
	do {
		if ((KeyCode = WaitForKeyCode ()) != 0) {
			if (CodesToKeys[KeyCode] == 0) {
				DBG ("not found (%x), buf: %s\n", KeyCode, buf);
				continue;
			}
				/* are we delaying an autorepeat? */
			if (LastKeyCode == KeyCode) {
					/* timers function as one-shot. */
				if (TimeElapsed (&RepeatTimer) == FALSE) continue;
			} else {
				LastKeyCode = KeyCode;
					/* lag autorepeat to avoid double entries. */
				SetElapsed (AutoDelay10MS, &RepeatTimer);
			}
			SysKeypress (ui_fd, CodesToKeys[KeyCode], 1);
			SysKeypress (ui_fd, CodesToKeys[KeyCode], 0);
			DBG ("keycode: %x\n", KeyCode);
		}
	} while (true);
	DBG ("destroyed!\n");
	DestroyUinputDevice(ui_fd);
	return 0;
}

